
(* ************************************************** *)
(* Noen basisfunksjoner *)

structure BasisFunksjoner =
struct
fun erBlank c = (c = #" " orelse c = #"\t" orelse c = #"\n" orelse c = #"\r")
fun erSiffer c = (#"0" <= c andalso c <= #"9")
fun erLiten c = (#"a" <= c andalso c <= #"z")
fun erStor c = (#"A" <= c andalso c <= #"Z")
fun erBokstav c = (erLiten c orelse erStor c)
fun erAlfanumerisk c = (erBokstav c orelse erSiffer c)
fun charToInt c = ord c - ord #"0";
fun charToStr c = Char.toString c;
fun intToStr c = Int.toString c;
fun prompt (str) =
    (print str;
     let open TextIO
	 val str = inputLine stdIn
     in if (str = "\n" orelse str = "") then inputLine stdIn
	else str
     end);
fun skrivTall n = print (intToStr(n) ^ "\n");
fun lesFil (filnavn : string): string =
    let open TextIO
	val innfil = openIn filnavn
	val str = inputAll innfil
    in closeIn innfil; str end
end;

open BasisFunksjoner;


(* ************************************************** *)
(* ENKEL SCANNER FOR L *)

(*  variabel -> v siffer  *)
(*  siffer -> 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9  *)
(*  tall -> siffer+  *)

exception Skannerfeil of string;
datatype token = Lik 
	       | Pluss
	       | Minus
	       | Input
	       | Output
	       | Var of int
	       | Tall of int;

fun ignorerLinje [] = []
  | ignorerLinje (#"\n" :: cs) = cs
  | ignorerLinje (c :: cs) = ignorerLinje cs;

fun scanTall (result, cs) =
    case cs of c :: cr =>
	       if erSiffer c
	       then scanTall(charToInt(c) + (10*result), cr)
	       else (result, cs)
	     | [] => (result, []);

fun lesTall (str) = #1(scanTall(0, (explode str)));


fun scan s =
    let fun scanh [] = []
	  | scanh (#"+" :: resten) = Pluss :: scanh resten 
	  | scanh (#"-" :: resten) = Minus :: scanh resten 
	  | scanh (#"=" :: resten) = Lik :: scanh resten 
	  | scanh (#"?" :: resten) = Input :: scanh resten 
	  | scanh (#"!" :: resten) = Output :: scanh resten 
	  | scanh (#"%" :: resten) = scanh (ignorerLinje(resten))
	  | scanh (#"v" :: siffer :: resten) =
	    if (erSiffer siffer) then Var(charToInt(siffer)) :: scanh resten
	    else raise Skannerfeil("v etterfulgt av " ^ charToStr(siffer))
	  | scanh (c :: resten) =
	    if (erBlank c)
	    then scanh resten
	    else
		if (erSiffer c)
		then let val (tall, rest) = scanTall (charToInt(c), resten)
		     in Tall(tall) :: scanh rest end
		else
		    raise Skannerfeil("Gjenkjenner ikke " ^ charToStr(c));
    in scanh (explode s) end;



(* ************************************************** *)
(* ENKEL INTERPRET FOR L *)

local
    open Array
    val Lager = array (10, 0)
in
fun setLager(i, x) = update(Lager, i, x);
fun getLager(i) = sub(Lager, i);
end;

fun lesTilLager (i) = setLager(i, lesTall(prompt("Gi verdi for v" ^ intToStr(i) ^ ": ")));
fun lesFraLager (i) = print ("v" ^ intToStr(i) ^ " = " ^ intToStr(getLager(i)) ^ "\n");


exception Parserfeil of string;
local 
fun program ts = setningsListe ts
and setningsListe ts = 
    case ts of Var(x) :: tr => setningsListeX (setning ts)
	     | Input  :: tr => setningsListeX (setning ts)
	     | Output :: tr => setningsListeX (setning ts)
	     | _ => raise Parserfeil("setningsListe: m begynne med setning ('v', '?' eller '!')")
and setningsListeX ts = 
    case ts of Var(x) :: tr => setningsListeX (setning ts)
	     | Input  :: tr => setningsListeX (setning ts)
	     | Output :: tr => setningsListeX (setning ts)
	     | _ => ts
and setning ts = 
    case ts of Var(x) :: tr => tilordning ts
	     | Input  :: tr => inputSetning ts
	     | Output :: tr => outputSetning ts
	     | _ => raise Parserfeil("setning: m begynne med 'v', '?' eller '!'")
and tilordning ts = 
    case ts of Var(x) :: tr =>
	       let val ts1 = (case tr of Lik :: rest => rest
				       | _ => raise Parserfeil("tilordning: 'v' m etterflges  av '='"))
		   val (ts2, tall1) = (case ts1 of Var(x) :: rest => (rest, getLager(x))
					| _ => raise Parserfeil("tilordning: '=' m etterflges  av 'v'"))
		   val (ts3, opr) = operator ts2
		   val (ts4, tall2) = operand ts3
	       in (setLager(x, opr(tall1, tall2)); ts4) end
	     | _ => raise Parserfeil("tilordning: m begynne med 'v'")
and operator ts =
    case ts of Pluss :: tr => (tr, op+)
	     | Minus :: tr => (tr, op-)
	     | _ => raise Parserfeil("operator: m begynne med '+' eller '-'")
and operand ts =
    case ts of Var(x)  :: tr => (tr, getLager(x))
	     | Tall(x) :: tr => (tr, x)
	     | _ => raise Parserfeil("operand: m begynne med 'v' eller tall")
and inputSetning ts =
    case ts of Input :: tr => (case tr of Var(x) :: rest => (lesTilLager(x);rest)
					| _ => raise Parserfeil("variabel: ikke paa formen vx"))
	     | _ => raise Parserfeil("inputSetning: m begynne med '?'")
and outputSetning ts =
    case ts of Output :: tr => (case tr of Var(x) :: rest => (lesFraLager(x);rest)
					| _ => raise Parserfeil("variabel: ikke paa formen vx"))
	     | _ => raise Parserfeil("outputSetning: m begynne med '?'");
in
fun interpret ts = 
    let val tr = program ts
    in if tr = []
       then ((print "Interpretering vellykket!\n"); tr)
       else raise Parserfeil("Tokens er igjen!")
    end
end;



(* ************************************************** *)
(* Funksjoner for  scanne og parse fra fil *)

fun feilhandler (Skannerfeil(str)) = (print ("Skannerfeil ved lesning av: " ^ str ^ "\n"); [])
  | feilhandler (Parserfeil(str)) = (print ("Parserfeil: " ^ str ^ "\n"); [])
  | feilhandler x = raise x;

fun scanFil filnavn = scan(lesFil(filnavn)) handle x => feilhandler(x);

fun interpretFil navn = (print ("Interpreterer " ^ navn ^ ": \n"); interpret(scanFil navn) handle x => feilhandler(x));

